# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.


import extra_type as new_pkg_type


# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

def convert(arg):
	if arg is None:
		return None

	if isinstance(arg, basestring) or isinstance(arg, int) or isinstance(arg, long) or isinstance(arg, float):
		return float(arg)
	
	raise TypeError("Unknown argument type")

#Validation Functions
def validate_positive(arg):
	if arg is None:
		return

	if arg <=0:
		raise ValueError("value is not positive")

def validate_nonnegative(arg):
	if arg is None:
		return

	if arg < 0:
		raise ValueError("value is negative")

def validate_range(arg, lower, upper):
	if arg is None:
		return

	if arg < lower or arg > upper:
		raise ValueError("not in required range [%f, %f]" %(float(lower), float(upper)))

#Formatting functions
def default_formatter(arg, **kwarg):
	return str(arg)


# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

from iotile.core.utilities.typedargs import type_system
from iotile.core.utilities import typedargs
from iotile.core.exceptions import *

def test_simplereturntype():
	@typedargs.return_type("string")
	def returns_string():
		return "hello"

	val = returns_string()
	formed = type_system.format_return_value(returns_string, val)

	assert formed == "hello"

def test_complexreturntype():
	@typedargs.return_type("map(string, integer)")
	def returns_map():
		return {"hello": 5}

	val = returns_map()
	formed = type_system.format_return_value(returns_map, val)

	assert formed == "hello: 5"



from iotile.core.exceptions import ValidationError
from .verifier import Verifier


class EnumVerifier(Verifier):
    """Verify that an object is one of a list of literal options.

    Args:
        desc (string): A description of the verifier
    """

    def __init__(self, options, desc=None):
        super(EnumVerifier, self).__init__(desc)
        self.options = options

    def verify(self, obj):
        """Verify that the object conforms to this verifier's schema.

        Args:
            obj (object): A python object to verify

        Raises:
            ValidationError: If there is a problem verifying the object, a
                ValidationError is thrown with at least the reason key set indicating
                the reason for the lack of validation.
        """

        if obj not in self.options:
            raise ValidationError("Object is not in list of enumerated options", reason='not in list of enumerated options', object=obj, options=self.options)

        return obj

    def format(self, indent_level, indent_size=4):
        """Format this verifier.

        Returns:
            string: A formatted string
        """

        name = self.format_name('Enumeration', indent_size)
        return self.wrap_lines(name, indent_level, indent_size)



"""State machine for parsing IOTile reports coming in on a streaming basis
"""

import pkg_resources
from iotile.core.exceptions import ArgumentError

class IOTileReportParser (object):
    """Accumulates data from a stream and emits IOTileReports

    Every time new data is available on the stream, add_data should be called.
    Every time a complete report has been received, the optional callback passed in will
    be called with an IOTileReport subclass.

    Args:
        report_callback (callable): A function to be called every time a new report is received
            The signature should be bool report_callback(report, context).  The return value is True to
            indicate that IOTileReportParser should also keep a copy of the report or
            False to indicate it should delete it.
        error_callback (callable): A function to be called every time an error occurs.
            The signature should be error_callback(error_code, message, context).  If a fatal
            error occurs, further parsing of reports will be stopped.
    """

    #States for parser state machine
    WaitingForReportType = 1
    WaitingForReportHeader = 2
    WaitingForCompleteReport = 3
    ErrorState = 4

    #Errors
    ErrorFindingReportType = 1
    ErrorParsingReportHeader = 2
    ErrorParsingCompleteReport = 3

    def __init__(self, report_callback=None, error_callback=None):
        self.report_callback = report_callback
        self.error_callback = error_callback

        self.raw_data = bytearray()
        self.state = IOTileReportParser.WaitingForReportType

        self.current_type = 0
        self.current_report_parser = None
        self.current_header_size = 0
        self.current_report_size = 0
        self.context = None

        self.known_formats = self._build_type_map()
        self.reports = []

    def add_data(self, data):
        """Add data to our stream, emitting reports as each new one is seen

        Args:
            data (bytearray): A chunk of new data to add
        """

        if self.state == self.ErrorState:
            return

        self.raw_data += bytearray(data)

        still_processing = True
        while still_processing:
            still_processing = self.process_data()

    def process_data(self):
        """Attempt to extract a report from the current data stream contents

        Returns:
            bool: True if further processing is required and process_data should be
                called again.
        """

        further_processing = False

        if self.state == self.WaitingForReportType and len(self.raw_data) > 0:
            self.current_type = self.raw_data[0]

            try:
                self.current_header_size = self.calculate_header_size(self.current_type)
                self.state = self.WaitingForReportHeader
                further_processing = True
            except Exception as exc:
                self.state = self.ErrorState

                if self.error_callback:
                    self.error_callback(self.ErrorFindingReportType, str(exc), self.context)
                else:
                    raise

        if self.state == self.WaitingForReportHeader and len(self.raw_data) >= self.current_header_size:
            try:
                self.current_report_size = self.calculate_report_size(self.current_type, self.raw_data[:self.current_header_size])
                self.state = self.WaitingForCompleteReport
                further_processing = True
            except Exception as exc:
                self.state = self.ErrorState

                if self.error_callback:
                    self.error_callback(self.ErrorParsingReportHeader, str(exc), self.context)
                else:
                    raise

        if self.state == self.WaitingForCompleteReport and len(self.raw_data) >= self.current_report_size:
            try:
                report_data = self.raw_data[:self.current_report_size]
                self.raw_data = self.raw_data[self.current_report_size:]

                report = self.parse_report(self.current_type, report_data)
                self._handle_report(report)
                self.state = self.WaitingForReportType
                further_processing = True
            except Exception as exc:
                self.state = self.ErrorState

                if self.error_callback:
                    self.error_callback(self.ErrorParsingCompleteReport, str(exc), self.context)
                else:
                    raise

        return further_processing

    def calculate_header_size(self, current_type):
        """Determine the size of a report header given its report type
        """

        fmt = self.known_formats[current_type]
        return fmt.HeaderLength()

    def calculate_report_size(self, current_type, report_header):
        """Determine the size of a report given its type and header
        """

        fmt = self.known_formats[current_type]
        return fmt.ReportLength(report_header)

    def parse_report(self, current_type, report_data):
        """Parse a report into an IOTileReport subclass
        """

        fmt = self.known_formats[current_type]
        return fmt(report_data)

    def _handle_report(self, report):
        """Try to emit a report and possibly keep a copy of it
        """

        keep_report = True

        if self.report_callback is not None:
            keep_report = self.report_callback(report, self.context)

        if keep_report:
            self.reports.append(report)

    @classmethod
    def _build_type_map(cls):
        """Build a map of all of the known report format processors
        """

        formats = {}

        for entry in pkg_resources.iter_entry_points('iotile.report_format'):
            report_format = entry.load()
            formats[report_format.ReportType] = report_format

        return formats

    @classmethod
    def DeserializeReport(cls, serialized):
        """Deserialize a report that has been serialized by calling report.serialize()

        Args:
            serialized (dict): A serialized report object
        """

        type_map = cls._build_type_map()

        if serialized['report_format'] not in type_map:
            raise ArgumentError("Unknown report format in DeserializeReport", format=serialized['report_format'])

        report = type_map[serialized['report_format']](serialized['encoded_report'])
        report.received_time = serialized['received_time']

        return report



"""Virtual Tile for testing TileBasedVirtualDevice."""

from iotile.core.hw.virtual import VirtualTile, tile_rpc
from iotile.core.hw.proxy.proxy import TileBusProxyObject
from iotile.core.utilities.typedargs import param, return_type, context

@context("TestProxy")
class TestTileProxy(TileBusProxyObject):
    @param("arg_1", "integer")
    @param("arg_2", "integer")
    @return_type("integer")
    def add(self, arg_1, arg_2):
        """Add two numbers."""
        res, = self.rpc(0x80, 0x00, arg_1, arg_2, arg_format="LL", result_format="L")
        return res

    @return_type("integer")
    def count(self):
        """Add two numbers."""
        res, = self.rpc(0x80, 0x01, result_format="L")
        return res

    @classmethod
    def ModuleName(cls):
        return 'test01'


class TestTile(VirtualTile):
    def __init__(self, address, _args):
        super(TestTile, self).__init__(address, 'test01')
        self._counter = 0

        # Test worker creation
        self.create_worker(self._increment_counter, 1.0)

    def _increment_counter(self):
        self._counter += 1

    @tile_rpc(0x8000, "LL", "L")
    def add(self, arg_1, arg_2):
        """Add arg_1 and arg_2."""

        return [arg_1 + arg_2]

    @tile_rpc(0x8001, "", "L")
    def count(self):
        """Return the current counter value."""

        return [self._counter]



from iotile.core.hw.commands import RPCCommand
from iotile.core.hw.exceptions import *
from time import sleep
import proxy

class TileBusProxyPlugin (object):
	"""
	Proxy plugin objects are mixin style classes that add functionality modules to proxy objects
	"""

	def __init__(self, parent):
		if not isinstance(parent, proxy.TileBusProxyObject):
			raise ArgumentError("Attempting to initialize a TileBusProxyPlugin with an invalid parent object", parent=parent)

		self._proxy = parent

	def rpc(self, feature, cmd, *args, **kw):
		return self._proxy.rpc(feature, cmd, *args, **kw)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

from iotile.core.utilities import typedargs
from iotile.core.utilities.typedargs import type_system
import unittest
import pytest
import os.path
from iotile.core.exceptions import *

def test_type_injection():
	import extra_type_package.extra_type as typeobj

	assert not type_system.is_known_type("test_injected_type1")

	type_system.inject_type("test_injected_type1", typeobj)
	assert type_system.is_known_type("test_injected_type1")

def test_external_module_injection():
	"""
	Test type injection from an external python module
	"""

	path = os.path.join(os.path.dirname(__file__), 'extra_types')

	assert not type_system.is_known_type('new_type')
	type_system.load_external_types(path)
	assert type_system.is_known_type('new_type')

def test_external_package_injection():
	"""
	Test type injection from an external python package
	"""

	path = os.path.join(os.path.dirname(__file__), 'extra_type_package')

	assert not type_system.is_known_type('new_pkg_type')
	type_system.load_external_types(path)
	assert type_system.is_known_type('new_pkg_type')

def test_external_package_injection_failure():
	"""
	Test type injection raises error from nonexistant path
	"""

	with pytest.raises(ArgumentError):
		path = os.path.join(os.path.dirname(__file__), 'extra_type_package_nonexistant')
		type_system.load_external_types(path)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

#annotate.py
from decorator import decorator
from iotile.core.exceptions import *
import inspect
from typeinfo import type_system
from collections import namedtuple
import sys

class BasicContext(dict):
    pass

def _check_and_execute(f, *args, **kwargs):
    """
    Check the type of all parameters with type information, converting 
    as appropriate and then execute the function.
    """

    convargs = []
    spec = inspect.getargspec(f)

    #Convert and validate all arguments
    for i in xrange(0, len(args)):
        arg = spec.args[i]
        val = _process_arg(f, arg, args[i])
        convargs.append(val)

    convkw = {}
    for key, val in kwargs:
        convkw[key] = _process_arg(f, key, val)

    #Ensure that only IOTileException subclasses are passed by the caller
    try:
        retval = f(*convargs, **convkw)
    except IOTileException:
        raise
    except Exception as unknown:
        raise APIError(str(unknown.args)), None, sys.exc_info()[2]

    return retval

def _process_arg(f, arg, value):
    """
    Ensure that value is a valid argument for the named 
    parameter arg based on the annotated type and validator
    information.  Any errors are raised as either 
    ConversionError or ValidationError exceptions.
    """

    if arg in f.params:
        val = type_system.convert_to_type(value, f.types[arg])
    else:
        val = value

    #Run all of the validators
    try:
        if arg in f.valids:
            for valid in f.valids[arg]:
                valid[0](val, *valid[1])
    except (ValueError,TypeError) as e:
        raise ValidationError(e.args[0], argument=arg, value=val)

    return val

def _parse_validators(type, valids):
    """
    Given a list of validator names or n-tuples, map the name to 
    a validation function given the type and return a list of 
    validation function, arguments tuples
    """

    outvals = []

    for val in valids:
        if isinstance(val, basestring):
            args = []
        elif len(val) > 1:
            args = val[1:]
            val = val[0]
        else:
            raise ValueError("You must pass either a n-tuple or a string to define a validator")

        name = "validate_%s" % str(val)

        if not hasattr(type, name):
            raise ValidationError("Cannot resolve validator", typename=str(type), method_name=name, methods=dir(type))

        func = getattr(type, name)
        outvals.append((func, args))

    return outvals

def get_spec(f):
    if inspect.isclass(f):
        f = f.__init__

    spec = inspect.getargspec(f)

    if spec.defaults is None:
        numreq = len(spec.args)
    else:
        numreq = len(spec.args) - len(spec.defaults)

    #If the first argument is self, don't return it
    start = 0
    if numreq > 0 and spec.args[0] == 'self':
        start = 1

    reqargs = spec.args[start:numreq]
    optargs = set(spec.args[numreq:])

    return reqargs, optargs

def spec_filled(req, opt, pos, kw):
    left = filter(lambda x: x not in kw, pos)
    left = req[len(left):]

    if len(left) == 0:
        return True

    return False

def get_signature(f):
    """
    Return the pretty signature for this function:
    foobar(type arg, type arg=val, ...)
    """

    name = f.__name__

    if inspect.isclass(f):
        f = f.__init__

    spec = inspect.getargspec(f)
    num_args = len(spec.args)

    num_def = 0
    if spec.defaults is not None:
        num_def = len(spec.defaults)
    
    num_no_def = num_args - num_def

    args = []
    for i in xrange(0, len(spec.args)):
        typestr = ""
        if i == 0 and spec.args[i] == 'self':
            continue

        if spec.args[i] in f.types:
            typestr = "%s " % f.types[spec.args[i]]

        if i >= num_no_def:
            default = str(spec.defaults[i-num_no_def])
            if len(default) == 0:
                default="''"

            args.append("%s%s=%s" % (typestr, str(spec.args[i]), default))
        else:
            args.append(typestr + str(spec.args[i]))

    return "%s(%s)" % (name, ", ".join(args))

def print_help(f):
    """
    Print usage information about a context or function.

    For contexts, just print the context name and its docstring
    For functions, print the function signature as well as its
    argument types.
    """

    if isinstance(f, BasicContext):
        name = context_name(f)

        print "\n" + name + "\n"
        doc = inspect.getdoc(f)
        if doc is not None:
            doc = inspect.cleandoc(doc)
            print doc

        return

    sig = get_signature(f)
    doc = inspect.getdoc(f)
    if doc is not None:
        doc = inspect.cleandoc(doc)

    print "\n" + sig + "\n"
    if doc is not None:
        print doc

    if inspect.isclass(f):
        f = f.__init__

    print "\nArguments:"
    for key in f.params.iterkeys():
        type = f.types[key]
        desc = ""
        if key in f.param_descs:
            desc = f.param_descs[key]

        print " - %s (%s): %s" % (key, type, desc)

def print_retval(f, value):
    if hasattr(f, 'typed_retval') and f.typed_retval == True:
        print(type_system.format_return_value(f, value))
        return

    if not hasattr(f, 'retval'):
        print(str(value))

    elif f.retval.printer[0] is not None:
        f.retval.printer[0](value)
    elif f.retval.desc != "":
        print("%s: %s" % (f.retval.desc, str(value)))
    else:
        print(str(value))

def find_all(container):
    if isinstance(container, dict):
        names = container.keys()
    else:
        names = dir(container)
    
    context = BasicContext()

    for name in names:
        #Ignore _ and __ names
        if name.startswith('_'):
            continue

        if isinstance(container, dict):
            obj = container[name]
        else:
            obj = getattr(container, name)

        #Check if this is an annotated object that should be included.  Check the type of
        #annotated to avoid issues with module imports where someone did from annotate import *
        #into the module causing an annotated symbol to be defined as a decorator
        
        #If we are in a dict context then strings point to lazily loaded modules so include them
        #too.
        if isinstance(container, dict) and isinstance(obj, basestring):
            context[name] = obj
        elif hasattr(obj, 'annotated') and isinstance(getattr(obj, 'annotated'), int):
            context[name] = obj

    return context

def context_from_module(module):
    """
    Given a module, create a context from all of the top level annotated
    symbols in that module.  
    """

    con = find_all(module)

    if hasattr(module, "__doc__"):
        setattr(con, "__doc__", module.__doc__)

    if hasattr(module, "_name_"):
        name = module._name_
    else:
        name = module.__name__

    setattr(con, '_annotated_name', name)
    setattr(con, 'context', True)

    con = annotated(con)

    return name, con


def check_returns_data(f):
    if hasattr(f, 'typed_retval') and f.typed_retval == True:
        return True

    if not hasattr(f, 'retval'):
        return False

    return f.retval.data

#Decorators
def param(name, type, *validators, **kwargs):
    def _param(f):
        f = annotated(f)

        f.params[name] = type_system.get_type(type)
        f.types[name] = type
        f.valids[name] = _parse_validators(f.params[name], validators)

        if 'desc' in kwargs:
            f.param_descs[name] = kwargs['desc']

        return decorator(_check_and_execute, f)

    return _param

def returns(desc=None, printer=None, data=True):
    """
    Specify how the return value of this function should be handled

    If data == True, then this function just returns data and does 
    not return a context so that the context for future calls remains
    unchanged.  
    """

    def _returns(f):
        annotated(f)

        f.retval = namedtuple("ReturnValue", ["desc", "printer", "data"])
        f.retval.desc = desc
        f.retval.printer = (printer,)
        f.retval.data = data

        return f

    return _returns

def stringable(f):
    """Specify that the return value for this function should just be printed as a string
    """

    f.retval = namedtuple("ReturnValue", ["desc", "printer", "data"])
    f.retval.desc = ""
    f.retval.printer = (None,)
    f.retval.data = True
    return f

def return_type(type, formatter=None):
    """
    Specify that this function returns a typed value

    type must be a type known to the MoMo type system and formatter
    must be a valid formatter for that type
    """

    def _returns(f):
        annotated(f)
        f.typed_retval = True
        f.retval_type = type_system.get_type(type)
        f.retval_typename = type
        f.retval_formatter = formatter

        return f

    return _returns

def context(name=None):
    """
    Declare that a class defines a MoMo context for use with the momo function for discovering
    and using functionality.
    """

    def _context(cls):
        annotated(cls)
        cls.context = True

        if name is not None:
            cls._annotated_name = name
        else:
            cls._annotated_name = cls.__name__

        return cls

    return _context

def finalizer(f):
    """
    Indicate that this function destroys the context in which it is invoked, such as a quit method
    on a subprocess or a delete method on an object.
    """

    f = annotated(f)
    f.finalizer = True
    return f

def context_name(context):
    """
    Given a context, return its proper name
    """

    if hasattr(context, "_annotated_name"):
        return context._annotated_name
    elif inspect.isclass(context):
        return context.__class__.__name__

    return str(context)

def takes_cmdline(f):
    f = annotated(f)
    f.takes_cmdline = True

    return f
    
def annotated(f):
    if not hasattr(f, 'params'):
        f.params = {}
    if not hasattr(f, 'valids'):
        f.valids = {}
    if not hasattr(f, 'types'):
        f.types = {}
    if not hasattr(f, 'param_descs'):
        f.param_descs = {}

    f.annotated = True
    f.finalizer = False
    f.takes_cmdline = False
    return f

def short_description(f):
    """
    Given an object with a docstring, return the first line of the docstring
    """

    doc = inspect.getdoc(f)
    if doc is not None:
        doc = inspect.cleandoc(doc)
        lines = doc.splitlines()
        return lines[0]

    return ""



import pkg_resources
import json
import os
import traceback
import imp
import inspect
import time
from adapter import DeviceAdapter
from iotile.core.exceptions import ArgumentError
from iotile.core.hw.virtual.virtualdevice import RPCInvalidIDError, TileNotFoundError, RPCNotFoundError, VirtualIOTileDevice


class VirtualAdapterAsyncChannel(object):
    """A channel for tracing and streaming data asynchronously from virtual devices
    """

    def __init__(self, adapter, iotile_id):
        self.adapter = adapter
        self.iotile_id = iotile_id

    def stream(self, report):
        """Queue data for streaming

        Args:
            report (IOTileReport): A report object to stream to a client
        """

        conn_id = self._find_connection(self.iotile_id)

        if conn_id is not None:
            self.adapter._trigger_callback('on_report', conn_id, report)

    def trace(self, data):
        """Queue data for tracing

        Args:
            data (bytearray, string): Unstructured data to trace to any
                connected client.
        """

        conn_id = self._find_connection(self.iotile_id)

        if conn_id is not None:
            self.adapter._trigger_callback('on_trace', conn_id, data)

    def _find_connection(self, iotile_id):
        """Find the connection corresponding to an iotile_id
        """

        for conn_id, dev in self.adapter.connections.iteritems():
            if dev.iotile_id == iotile_id:
                return conn_id

        return None


class VirtualDeviceAdapter(DeviceAdapter):
    """Callback based adapter that gives access to one or more virtual devices

    The adapter is created and serves access to the virtual_devices that are
    found by name in the entry_point group iotile.virtual_device.

    Args:
        port (string): A port description that should be in the form of
            device_name1@<optional_config_json1;device_name2@optional_config_json2
    """

    # Make devices expire after a long time only
    ExpirationTime = 600000

    def __init__(self, port):
        super(VirtualDeviceAdapter, self).__init__()

        devs = port.split(';')
        loaded_devs = {}

        # This needs to be initialized before any VirtualAdapterAsyncChannels are
        # created because those could reference it
        self.connections = {}

        for dev in devs:
            name, sep, config = dev.partition('@')

            if len(config) == 0:
                config = None

            loaded_dev = self._load_device(name, config)
            loaded_dev.start(VirtualAdapterAsyncChannel(self, loaded_dev.iotile_id))
            loaded_devs[loaded_dev.iotile_id] = loaded_dev

        self.devices = loaded_devs
        self.scan_interval = self.get_config('scan_interval', 1.0)
        self.last_scan = None
        
        self.set_config('probe_required', True)
        self.set_config('probe_supported', True)

    def _find_device_script(self, script_path):
        """Import a virtual device from a file rather than an installed module

        script_path must point to a python file ending in .py that contains exactly one
        VirtualIOTileDevice class definitions.  That class is loaded and executed as if it
        were installed.

        Args:
            script_path (string): The path to the script to load

        Returns:
            VirtualIOTileDevice: A subclass of VirtualIOTileDevice that was loaded from script_path
        """

        search_dir, filename = os.path.split(script_path)
        if search_dir == '':
            search_dir = './'

        if filename == '' or not os.path.exists(script_path):
            raise ArgumentError("Could not find script to load virtual device", path=script_path)

        module_name, ext = os.path.splitext(filename)
        if ext != '.py':
            raise ArgumentError("Script did not end with .py", filename=filename)

        try:
            file = None
            file, pathname, desc = imp.find_module(module_name, [search_dir])
            mod = imp.load_module(module_name, file, pathname, desc)
        finally:
            if file is not None:
                file.close()

        devs = filter(lambda x: inspect.isclass(x) and issubclass(x, VirtualIOTileDevice) and x != VirtualIOTileDevice, mod.__dict__.itervalues())
        if len(devs) == 0:
            raise ArgumentError("No VirtualIOTileDevice subclasses were defined in script", path=script_path)
        elif len(devs) > 1:
            raise ArgumentError("More than one VirtualIOTileDevice subclass was defined in script", path=script_path, devices=devs)

        return devs[0]

    def _load_device(self, name, config):
        """Load a device either from a script or from an installed module
        """

        if config is None:
            config_dict = {}
        else:
            try:
                with open(config, "rb") as conf:
                    data = json.load(conf)
            except IOError, exc:
                raise ArgumentError("Could not open config file", error=str(exc), path=config)

            if 'device' not in data:
                raise ArgumentError("Invalid configuration file passed to VirtualDeviceAdapter", device_name=name, config_path=config, missing_key='device')

            config_dict = data['device']

        if name.endswith('.py'):
            device_factory = self._find_device_script(name)
            return device_factory(config_dict)

        seen_names = []
        for entry in pkg_resources.iter_entry_points('iotile.virtual_device'):
            if entry.name == name:
                device_factory = entry.load()
                return device_factory(config_dict)

            seen_names.append(entry.name)

        raise ArgumentError("Could not find virtual_device by name", name=name, known_names=seen_names)

    def can_connect(self):
        """Can this adapter have another simultaneous connection

        There is no limit on the number of simultaneous virtual connections

        Returns:
            bool: whether another connection is allowed
        """

        return True

    def connect_async(self, connection_id, connection_string, callback):
        """Asynchronously connect to a device

        Args:
            connection_id (int): A unique identifer that will refer to this connection
            connection_string (string): A DeviceAdapter specific string that can be used to connect to
                a device using this DeviceAdapter.
            callback (callable): A function that will be called when the connection attempt finishes as
                callback(conection_id, adapter_id, success: bool, failure_reason: string or None)
        """

        id_number = int(connection_string)
        if id_number not in self.devices:
            if callback is not None:
                callback(connection_id, self.id, False, "could not find device to connect to")
            return

        if id_number in [x.iotile_id for x in self.connections.itervalues()]:
            if callback is not None:
                callback(connection_id, self.id, False, "device was already connected to")
            return

        dev = self.devices[id_number]
        dev.connected = True

        self.connections[connection_id] = dev

        if callback is not None:
            callback(connection_id, self.id, True, "")

    def disconnect_async(self, conn_id, callback):
        """Asynchronously disconnect from a connected device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        dev.connected = False
        del self.connections[conn_id]

        if callback is not None:
            callback(conn_id, self.id, True, "")

    def _open_rpc_interface(self, conn_id, callback):
        """Open the RPC interface on a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        dev.open_rpc_interface()

        if callback is not None:
            callback(conn_id, self.id, True, "")

    def _open_streaming_interface(self, conn_id, callback):
        """Open the streaming interface on a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        reports = dev.open_streaming_interface()

        if callback is not None:
            callback(conn_id, self.id, True, "")

        for report in reports:
            self._trigger_callback('on_report', conn_id, report)

    def _open_tracing_interface(self, conn_id, callback):
        """Open the tracing interface on a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")

            return

        dev = self.connections[conn_id]
        traces = dev.open_tracing_interface()

        if callback is not None:
            callback(conn_id, self.id, True, "")

        for trace in traces:
            self._trigger_callback('on_trace', conn_id, trace)

    def _open_script_interface(self, conn_id, callback):
        """Open the script interface on a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        dev.open_script_interface()

        if callback is not None:
            callback(conn_id, self.id, True, "")

    def send_rpc_async(self, conn_id, address, rpc_id, payload, timeout, callback):
        """Asynchronously send an RPC to this IOTile device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            address (int): the addres of the tile that we wish to send the RPC to
            rpc_id (int): the 16-bit id of the RPC we want to call
            payload (bytearray): the payload of the command
            timeout (float): the number of seconds to wait for the RPC to execute
            callback (callable): A callback for when we have finished the RPC.  The callback will be called as"
                callback(connection_id, adapter_id, success, failure_reason, status, payload)
                'connection_id': the connection id
                'adapter_id': this adapter's id
                'success': a bool indicating whether we received a response to our attempted RPC
                'failure_reason': a string with the reason for the failure if success == False
                'status': the one byte status code returned for the RPC if success == True else None
                'payload': a bytearray with the payload returned by RPC if success == True else None
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, 'Device is not in connected state', None, None)
            return

        dev = self.connections[conn_id]

        status = (1 << 6)
        try:
            response = dev.call_rpc(address, rpc_id, str(payload))
            if len(response) > 0:
                status |= (1 << 7)
        except (RPCInvalidIDError, RPCNotFoundError):
            status = 2
            response = ""
        except TileNotFoundError:
            status = 0xFF
            response = ""
        except Exception:
            #Don't allow exceptions or we will deadlock
            status = 3
            response = ""

            print("*** EXCEPTION OCCURRED IN RPC ***")
            traceback.print_exc()
            print("*** END EXCEPTION ***")

        response = bytearray(response)
        callback(conn_id, self.id, True, "", status, response)

    def send_script_async(self, conn_id, data, progress_callback, callback):
        """Asynchronously send a a script to this IOTile device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            data (string): the script to send to the device
            progress_callback (callable): A function to be called with status on our progress, called as:
                progress_callback(done_count, total_count)
            callback (callable): A callback for when we have finished sending the script.  The callback will be called as"
                callback(connection_id, adapter_id, success, failure_reason)
                'connection_id': the connection id
                'adapter_id': this adapter's id
                'success': a bool indicating whether we received a response to our attempted RPC
                'failure_reason': a string with the reason for the failure if success == False
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, 'Device is not in connected state')
            return

        # Simulate some progress callbacks (0, 50%, 100%)
        progress_callback(0, len(data))
        dev = self.connections[conn_id]
        dev.script = data
        progress_callback(len(data)//2, len(data))
        progress_callback(len(data), len(data))

        callback(conn_id, self.id, True, None)

    def _send_scan_event(self, device):
        """Send a scan event from a device
        """

        info = {
            'connection_string': str(device.iotile_id),
            'uuid': device.iotile_id,
            'signal_strength': 100
        }
        self._trigger_callback('on_scan', self.id, info, self.ExpirationTime)

    def periodic_callback(self):
        if self.last_scan is None or time.time() > (self.scan_interval + self.last_scan):
            for dev in self.devices.itervalues():
                self._send_scan_event(dev)

    def stop_sync(self):
        for dev in self.devices.itervalues():
            dev.stop()

    def probe_async(self, callback):
        """Send advertisements for all connected virtual devices.

        Args:
            callback (callable): A callback for when the probe operation has completed.
                callback should have signature callback(adapter_id, success, failure_reason) where:
                    success: bool
                    failure_reason: None if success is True, otherwise a reason for why we could not probe
        """

        for dev in self.devices.itervalues():
                self._send_scan_event(dev)

        callback(self.id, True, None)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

import sys
import os
import shlex

from iotile.core.utilities.typedargs.shell import HierarchicalShell, posix_lex
from iotile.core.exceptions import *
from iotile.core.utilities.typedargs import annotate, type_system
from iotile.core.utilities.rcfile import RCFile
from iotile.core.dev.registry import ComponentRegistry
import iotile.core.hw.transport.cmdstream as cmdstream

from multiprocessing import freeze_support
import traceback

def main():
    type_system.interactive = True
    line = sys.argv[1:]

    norc = False
    if len(line) > 0 and line[0] == '--norc':
        norc = True
        line = line[1:]

    if len(line) > 0 and line[0] == '--rcfile':
        rc = RCFile('iotile')
        print rc.path
        return 0

    shell = HierarchicalShell('iotile', no_rc=norc)
        
    shell.root_add("registry", "iotile.core.dev.annotated_registry,registry")
    shell.root_add("config", "iotile.core.dev.config,ConfigManager")
    shell.root_add('hw', "iotile.core.hw.hwmanager,HardwareManager")

    reg = ComponentRegistry()
    plugins = reg.list_plugins()
    for k, v in plugins.iteritems():
        shell.root_add(k, v)

    finished = False

    try:
        while len(line) > 0:
            line, finished = shell.invoke(line)
    except APIError:
        traceback.print_exc()
        cmdstream.do_final_close()
        return 1
    except IOTileException as e:
        print e.format()
        #if the command passed on the command line fails, then we should
        #just exit rather than drop the user into a shell.
        cmdstream.do_final_close()
        return 1
    except Exception:
        #Catch all exceptions because otherwise we won't properly close cmdstreams
        #since the program will be said to except 'abnormally'
        traceback.print_exc()
        cmdstream.do_final_close()
        return 1

    if finished:
        cmdstream.do_final_close()
        return 0

    #Setup file path and function name completion
    #Handle special case of importing pyreadline on Windows
    #See: http://stackoverflow.com/questions/6024952/readline-functionality-on-windows-with-python-2-7
    import glob
    try:
        import readline
    except ImportError:
        import pyreadline as readline

    def complete(text, state):
        buf = readline.get_line_buffer()
        if buf.startswith('help ') or " " not in buf:
            funcs = shell.valid_identifiers()
            return filter(lambda x: x.startswith(text), funcs)[state]

        return (glob.glob(os.path.expanduser(text)+'*')+[None])[state]

    readline.set_completer_delims(' \t\n;')
    #Handle Mac OS X special libedit based readline
    #See: http://stackoverflow.com/questions/7116038/python-tab-completion-mac-osx-10-7-lion
    if readline.__doc__ is not None and 'libedit' in readline.__doc__:
        readline.parse_and_bind("bind ^I rl_complete")
    else:
        readline.parse_and_bind("tab: complete")

    readline.set_completer(complete)

    #We ended the initial command with a context, start a shell
    try:
        while True:
            linebuf = raw_input("(%s) " % shell.context_name())

            # Skip comments automatically
            if len(linebuf) > 0 and linebuf[0] == '#':
                continue

            line = shlex.split(linebuf, posix=posix_lex)

            #Automatically remove enclosing double quotes on windows since they are not removed by shlex in nonposix mode
            def remove_quotes(x):
                if len(x) > 0 and x.startswith(("'", '"')) and x[0] == x[-1]:
                    return x[1:-1]

                return x

            if not posix_lex:
                line = map(remove_quotes, line)

            #Catch exception outside the loop so we stop invoking submethods if a parent
            #fails because then the context and results would be unpredictable
            try:
                while len(line) > 0:
                    line, finished = shell.invoke(line)
            except APIError as e:
                traceback.print_exc()
            except IOTileException as e:
                print e.format()

            if shell.finished():
                break

    #Make sure to catch ^C and ^D so that we can cleanly dispose of subprocess resources if 
    #there are any.
    except EOFError:
        print ""
    except KeyboardInterrupt:
        print ""
    finally:
        #Make sure we close any open CMDStream communication channels so that we don't lockup at exit
        cmdstream.do_final_close()



import struct
import sys

def unpack(fmt, arg):
	"""
	A shim around struct.unpack to allow it to work on python 2.7.3.
	"""

	if isinstance(arg, bytearray) and not (sys.version_info >= (2,7,5)):
		return struct.unpack(fmt, str(arg))

	return struct.unpack(fmt, arg)



"""Mock IOTile device class for testing other components interactions with IOTile devices
"""

import inspect
from iotile.core.exceptions import InternalError, ArgumentError
from iotile.core.hw.reports.individual_format import IndividualReadingReport
from iotile.core.hw.reports.report import IOTileReading
from .base_runnable import BaseRunnable
from .common_types import RPCInvalidIDError, RPCNotFoundError, TileNotFoundError, RPCDispatcher, rpc  # pylint: disable=W0611; rpc import needed for backwards compatibility


# pylint: disable=R0902,R0904; backwards compatibility methods and properties already referenced in other modules
class VirtualIOTileDevice(BaseRunnable):
    """A Virtual IOTile device that can be interacted with as if it were a real one.

    This is the base class of all other Virtual IOTile devices.  It allows defining
    RPCs directly using decorators and the `register_rpc` function.  Subclasses
    implement virtual tiles that modularize building complex virtual devices from
    reusable pieces.

    This class also implements the required controller status RPC that allows
    matching it with a proxy object.

    Args:
        iotile_id (int): A 32-bit integer that specifies the globally unique ID
            for this IOTile device.
        name (string): The 6 byte name that should be returned when anyone asks
            for the controller's name of this IOTile device using an RPC
    """

    def __init__(self, iotile_id, name):
        super(VirtualIOTileDevice, self).__init__()

        self._rpc_overlays = {}
        self._tiles = {}

        self.name = name
        self.iotile_id = iotile_id
        self.reports = []
        self.traces = []
        self.script = bytearray()

        self._interface_status = {}
        self._interface_status['connected'] = False
        self._interface_status['stream'] = False
        self._interface_status['trace'] = False
        self._interface_status['script'] = False
        self._interface_status['rpc'] = False

        # For this device to push streams or tracing data through a VirtualInterface, it
        # needs access to that interface's push channel
        self._push_channel = None

        # Iterate through all of our member functions and see the ones that are
        # RPCS and add them to the RPC handler table
        for _name, value in inspect.getmembers(self, predicate=inspect.ismethod):
            if hasattr(value, 'is_rpc'):
                self.register_rpc(value.rpc_addr, value.rpc_id, value)

    @property
    def connected(self):
        """Whether someone is connected to the virtual device."""

        return self._interface_status['connected']

    @connected.setter
    def connected(self, value):
        self._interface_status['connected'] = value

    @property
    def stream_iface_open(self):
        """Whether the streaming interface is opened.

        Deprecated:
            3.14.5: Use interface_open('stream') instead
        """

        return self._interface_status['stream']

    @property
    def trace_iface_open(self):
        """Whether the tracing interface is opened.

        Deprecated:
            3.14.5: Use interface_open('trace') instead
        """

        return self._interface_status['trace']

    @property
    def pending_data(self):
        """Whether there are streaming reports waiting to be sent."""

        return len(self.reports) > 0

    def start(self, channel):
        """Start running this virtual device including any necessary worker threads.

        Args:
            channel (IOTilePushChannel): the channel with a stream and trace
                routine for streaming and tracing data through a VirtualInterface
        """

        if self._started:
            raise InternalError("The method start() was called twice on VirtualIOTileDevice.")

        self._push_channel = channel
        self.start_workers()

    def stop(self):
        """Stop running this virtual device including any worker threads."""

        self.stop_workers()

    def stream(self, report):
        """Stream a report asynchronously.

        If no one is listening for the report, the report may be dropped,
        otherwise it will be queued for sending

        Args:
            report (IOTileReport): The report that should be streamed
        """

        if self._push_channel is None:
            return

        self._push_channel.stream(report)

    def stream_realtime(self, stream, value):
        """Stream a realtime value as an IndividualReadingReport.

        If the streaming interface of the VirtualInterface this
        VirtualDevice is attached to is not opened, the realtime
        reading may be dropped.

        Args:
            stream (int): The stream id to send
            value (int): The stream value to send
        """

        if not self.stream_iface_open:
            return

        reading = IOTileReading(0, stream, value)

        report = IndividualReadingReport.FromReadings(self.iotile_id, [reading])
        self.stream(report)

    def trace(self, data):
        """Trace data asynchronously.

        If no one is listening for traced data, it will be dropped
        otherwise it will be queued for sending.

        Args:
            data (bytearray, string): Unstructured data to trace to any
                connected client.
        """

        if self._push_channel is None:
            return

        self._push_channel.trace(data)

    def register_rpc(self, address, rpc_id, func):
        """Register a single RPC handler with the given info.

        This function can be used to directly register individual RPCs,
        rather than delegating all RPCs at a given address to a virtual
        Tile.

        If calls to this function are mixed with calls to add_tile for
        the same address, these RPCs will take precedence over what is
        defined in the tiles.

        Args:
            address (int): The address of the mock tile this RPC is for
            rpc_id (int): The number of the RPC
            func (callable): The function that should be called to handle the
                RPC.  func is called as func(payload) and must return a single
                string object of up to 20 bytes with its response
        """

        if rpc_id < 0 or rpc_id > 0xFFFF:
            raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

        if address not in self._rpc_overlays:
            self._rpc_overlays[address] = RPCDispatcher()

        self._rpc_overlays[address].add_rpc(rpc_id, func)

    def call_rpc(self, address, rpc_id, payload=""):
        """Call an RPC by its address and ID.

        Args:
            address (int): The address of the mock tile this RPC is for
            rpc_id (int): The number of the RPC
            payload (str): A byte string of payload parameters up to 20 bytes

        Returns:
            str: The response payload from the RPC
        """

        if rpc_id < 0 or rpc_id > 0xFFFF:
            raise RPCInvalidIDError("Invalid RPC ID: {}".format(rpc_id))

        if address not in self._rpc_overlays and address not in self._tiles:
            raise TileNotFoundError("Unknown tile address, no registered handler", address=address)

        overlay = self._rpc_overlays.get(address, None)
        tile = self._tiles.get(address, None)
        if overlay is not None and overlay.has_rpc(rpc_id):
            return overlay.call_rpc(rpc_id, payload)
        elif tile is not None and tile.has_rpc(rpc_id):
            return tile.call_rpc(rpc_id, payload)

        raise RPCNotFoundError("Could not find RPC 0x%X at address %d" % (rpc_id, address))

    def add_tile(self, address, tile):
        """Add a tile to handle all RPCs at a given address.

        Args:
            address (int): The address of the tile
            tile (RPCDispatcher): A tile object that inherits from RPCDispatcher
        """

        if address in self._tiles:
            raise ArgumentError("Tried to add two tiles at the same address", address=address)

        self._tiles[address] = tile

    def open_rpc_interface(self):
        """Called when someone opens an RPC interface to the device."""

        self._interface_status['rpc'] = True

    def close_rpc_interface(self):
        """Called when someone closes an RPC interface to the device."""

        self._interface_status['rpc'] = False

    def open_script_interface(self):
        """Called when someone opens a script interface on this device."""

        self._interface_status['script'] = True

    def close_script_interface(self):
        """Called when someone closes a script interface on this device."""

        self._interface_status['script'] = False

    def open_streaming_interface(self):
        """Called when someone opens a streaming interface to the device.

        Returns:
            list: A list of IOTileReport objects that should be sent out
                the streaming interface.
        """

        self._interface_status['stream'] = True
        return self.reports

    def close_streaming_interface(self):
        """Called when someone closes the streaming interface to the device."""

        self._interface_status['stream'] = False

    def open_tracing_interface(self):
        """Called when someone opens a tracing interface to the device.

        Returns:
            list: A list of bytearray objects that should be sent out
                the tracing interface.
        """

        self._interface_status['trace'] = True
        return self.traces

    def close_tracing_interface(self):
        """Called when someone closes the tracing interface to the device."""

        self._interface_status['trace'] = True

    def push_script_chunk(self, chunk):
        """Called when someone pushes a new bit of a TRUB script to this device

        Args:
            chunk (str): a buffer with the next bit of script to append
        """

        self.script += bytearray(chunk)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

#shell.py
#Given a command line string, attempt to map it to a function and fill in 
#the parameters based on that function's annotated type information.

from iotile.core.exceptions import *
import annotate
import inspect
import shlex
from typeinfo import type_system
from iotile.core.utilities.rcfile import RCFile
import os.path
import platform
import importlib

posix_lex = platform.system() != 'Windows'

builtin_help = {
    'help': "help [function]: print help information about the current context or a function",
    'back': "back: replace the current context with its parent",
    'quit': "quit: quit the momo shell"
}

def process_kwarg(flag, arg_it):
    flag = flag[2:]
    skip = 0

    #Check if of the form name=value
    if '=' in flag:
        name, value = flag.split('=')
    else:
        name = flag
        value = next(arg_it)
        skip = 1

    return name, value, skip

@annotate.param("package", "path", "exists", desc="Path to the python package containing the types")
@annotate.param("module", "string", desc="The name of the submodule to load from package, if applicable")
def import_types(package, module=None):
    """
    Add externally defined types from a python package or module

    The MoMo type system is built on the typedargs package, which defines what kinds of types
    can be used for function arguments as well as how those types should be displayed and 
    converted from binary representations or strings.  This function allows you to define external
    types in a separate package and import them into the MoMo type system so that they can be used.
    You might want to do this if you have custom firmware objects that you would like to interact with
    or that are returned in a syslog entry, for example.  

    All objects defined in the global namespace of package (if module is None) or package.module if 
    module is specified that define valid types will be imported and can from this point on be used
    just like any primitive type defined in typedargs itself.  Imported types are indistinguishable 
    from primivtive types like string, integer and path.
    """

    if module is None:
        path = package
    else:
        path = os.path.join(package, module)

    type_system.load_external_types(path)

def print_dir(context):
    doc = inspect.getdoc(context)
    
    print ""
    print annotate.context_name(context)

    if doc is not None:
        doc = inspect.cleandoc(doc)
        print doc
    
    print "\nDefined Functions:"

    if isinstance(context, dict):
        funs = context.keys()
    else:
        funs = annotate.find_all(context)
    
    for fun in funs:
        fun = find_function(context, fun)
        if isinstance(fun, annotate.BasicContext):
            print " - " + fun._annotated_name
        else:
            print " - " + annotate.get_signature(fun)

        if annotate.short_description(fun) != "":
            print "   " + annotate.short_description(fun) + '\n'
        else:
            print ""

    print "\nBuiltin Functions"
    for bi in builtin_help.values():
        print ' - ' + bi

    print ""

def print_help(context, fname):
    if fname in builtin_help:
        print builtin_help[fname]
        return

    func = find_function(context, fname)
    annotate.print_help(func)

def _do_help(context, line):
    args = line[1:]
    if len(args) == 0:
        print_dir(context)
    elif len(args) == 1:
        print_help(context, args[0])
    else:
        print "Too many arguments:", args
        print "Usage: help [function]"

    return [], True

def deferred_add(add_action):
    """
    Perform a lazy import of a context so that we don't have a huge initial startup time
    loading all of the modules that someone might want even though they probably only 
    will use a few of them.
    """

    module, sep, obj = add_action.partition(',')

    mod = importlib.import_module(module)
    if obj == "":
        name, con = annotate.context_from_module(mod)
        return con

    if hasattr(mod, obj):
        return getattr(mod, obj)

    raise ArgumentError("Attempted to import nonexistent object from module", module=module, object=obj)

def find_function(context, funname):
    func = None
    if isinstance(context, dict):
        if funname in context:
            func = context[funname]

            #Allowed lazy loading of functions
            if isinstance(func, basestring):
                func = deferred_add(func)
                context[funname] = func
    elif hasattr(context, funname):
        func = getattr(context, funname)

    if func is None:
        raise NotFoundError("Function not found", function=funname)

    return func

@annotate.context("root")
class InitialContext(dict):
    pass

class HierarchicalShell:
    def __init__(self, name, no_rc=False):
        """
        Create a new HierarchicalShell, optionally loading initialization
        statements from an RCFile based on the passed identifier 'name'.
        """

        self.name = name
        self.init_commands = {}

        if not no_rc:
            self._load_rc()

        self.root = InitialContext()
        self.contexts = [self.root]

        self.root_add('import_types', import_types)

        #Initialize the root context if required
        self._check_initialize_context()

    def root_update(self, dict_like):
        """
        Add all of the entries in the dictionary like object dict_like to the root
        context 
        """

        self.root.update(dict_like)

    def root_add(self, name, value):
        """
        Add a single function to the root context
        """

        self.root[name] = value

    def context_name(self):
        """
        Get the string name of the current context
        """

        return annotate.context_name(self.contexts[-1])

    def finished(self):
        """
        If there are no more contexts on the context list, then we cannot execute any
        commands and the shell has finished its useful life.
        """

        return len(self.contexts) == 0

    def valid_identifiers(self):
        """
        Return a list of the valid identifiers that can be called given the context that
        we are in.  Callers can use this result to perform autocomplete if they would like.
        """

        funcs = annotate.find_all(self.contexts[-1]).keys() + builtin_help.keys()
        return funcs

    def _load_rc(self):
        """
        Load context initialization commands from a configuration file.  The 
        file should have the format:
        [Context1]
        <list of commands>

        [OldContext1.Subcontext1]
        <list of commands>

        where list of commands is a list of lines that are executed using invoke
        whenever a context matching the identifier inside of the brackets is created
        for the first time.  This lets you run custom configuration routines to take 
        care of tiresome initialization calls automatically.
        """

        rcfile = RCFile(self.name)
        context = None
        cmds = []

        for line in rcfile.contents:
            line = line.strip()

            if len(line) == 0:
                continue

            if line[0] == '[':
                lasti = line.find(']')
                if lasti == -1:
                    raise InternalError("Syntax Error in rcfile, missing closing ]", line=line, name=self.name, path=rcfile.path)

                if context is not None:
                    self.init_commands[context] = cmds
                
                #Start a new context
                context = line[1:lasti].strip()
                cmds = []
            else:
                #Process a command line
                if context is None:
                    raise InternalError("Syntax Error in rcfile, command given before a context was specified", line=line, name=self.name, path=rcfile.path)

                cmds.append(line)

        if context is not None:
            self.init_commands[context] = cmds

    def _check_initialize_context(self):
        """
        Check if our context matches something that we have initialization commands
        for.  If so, run them to initialize the context before proceeding with other
        commands.
        """

        path = ".".join([annotate.context_name(x) for x in self.contexts])

        #Make sure we don't clutter up the output with return values from
        #initialization functions
        old_interactive = type_system.interactive
        type_system.interactive = False

        for key,cmds in self.init_commands.iteritems():
            if path.endswith(key):
                for cmd in cmds:
                    line = shlex.split(cmd, posix=posix_lex)

                    #Automatically remove enclosing double quotes on windows since they are not removed by shlex in nonposix mode
                    def remove_quotes(x):
                        if len(x) > 0 and x.startswith(("'", '"')) and x[0] == x[-1]:
                            return x[1:-1]

                        return x

                    if not posix_lex:
                        line = map(remove_quotes, line)

                    self.invoke(line)

        type_system.interactive = old_interactive

    def invoke(self, line):
        """
        Given a list of command line arguments, attempt to find the function being specified
        and map the passed arguments to that function based on its annotated type information. 
        """

        funname = line[0]
        context = self.contexts[-1]

        #Check if we are asked for help
        if funname == 'help':
            return _do_help(context, line)
        if funname == 'quit':
            del self.contexts[:]
            return [], True
        if funname == 'back':
            self.contexts.pop()
            return [], True

        func = find_function(context, funname)

        #If this is a context derived from a module or package, just jump to it
        #since there is no initialization function
        if isinstance(func, annotate.BasicContext):
            self.contexts.append(func)
            self._check_initialize_context()
            return line[1:], False

        #find out how many position and kw args this function takes
        posset,kwset = annotate.get_spec(func)

        #If the function wants arguments directly, do not parse them
        if func.takes_cmdline == True:
            val = func(line[1:])
        else:
            arg_it = (x for x in line[1:])
            kwargs = {}
            posargs = []

            i = 1
            for arg in arg_it:
                if arg.startswith('--') or ((arg.startswith('-') and len(arg)==2)):
                    name, val, skip = process_kwarg(arg, arg_it)
                    kwargs[name] = val
                    i += skip
                else:
                    if annotate.spec_filled(posset, kwset, posargs, kwargs):
                        break

                    posargs.append(arg)

                i += 1

            if not annotate.spec_filled(posset, kwset, posargs, kwargs):
                raise ArgumentError("too few arguments")

            val = func(*posargs, **kwargs)

        #Update our current context if this function destroyed it or returned a new one.
        finished = True

        if func.finalizer == True:
            self.contexts.pop()
        elif val is not None:
            if annotate.check_returns_data(func):
                if type_system.interactive:
                    annotate.print_retval(func, val)
            else:
                self.contexts.append(val)
                self._check_initialize_context()
                finished = False

        return line[i:], finished



"""IOTileReport subclass for readings packaged as individual readings
"""

import datetime
import struct
from .report import IOTileReport, IOTileReading
from iotile.core.utilities.packed import unpack
from iotile.core.exceptions import ArgumentError, NotFoundError, ExternalError
import iotile.core.hw.auth.auth_provider as auth_provider
from iotile.core.hw.auth.auth_chain import ChainedAuthProvider

class SignedListReport(IOTileReport):
    """A report that consists of a signed list of readings.

    Args:
        rawreport (bytearray): The raw data of this report
    """

    ReportType = 1

    def __init__(self, rawreport, **kwargs):
        super(SignedListReport, self).__init__(rawreport, signed=True, encrypted=False, **kwargs)

    @classmethod
    def HeaderLength(cls):
        """Return the length of a header needed to calculate this report's length

        Returns:
            int: the length of the needed report header
        """

        return 20

    @classmethod
    def ReportLength(cls, header):
        """Given a header of HeaderLength bytes, calculate the size of this report
        """

        first_word, = unpack("<L", header[:4])

        length = (first_word >> 8)
        return length

    @classmethod
    def FromReadings(cls, uuid, readings, signature_method=auth_provider.HashOnlySHA256Signature, signer=None, report_id=IOTileReading.InvalidReadingID, selector=0xFFFF, streamer=0, sent_timestamp=0):
        """Generate an instance of the report format from a list of readings and a uuid.

        The signed list report is created using the passed readings and signed using the specified method
        and AuthProvider.  If no auth provider is specified, the report is signed using the default authorization
        chain.

        Args:
            uuid (int): The uuid of the deviec that this report came from
            readings (list): A list of IOTileReading objects containing the data in the report
            signature_method (int): The method that should be used to sign the report (must be supported
                by an auth_provider)
            signer (AuthProvider): An optional preconfigured AuthProvider that should be used to sign this
                report.  If no AuthProvider is provided, the default ChainedAuthProvider is used.
            report_id (int): The id of the report.  If not provided it defaults to IOTileReading.InvalidReadingID.
                Note that you can specify anything you want for the report id but for actual IOTile devices
                the report id will always be greater than the id of all of the readings contained in the report
                since devices generate ids sequentially.
            selector (int): The streamer selector of this report.  This can be anything but if the report came from
                a device, it would correspond with the query the device used to pick readings to go into the report.
            streamer (int): The streamer id that this reading was sent from.
            sent_timestamp (int): The device's uptime that sent this report.
        """

        lowest_id = IOTileReading.InvalidReadingID
        highest_id = IOTileReading.InvalidReadingID

        report_len = 20 + 16*len(readings) + 24
        len_low = report_len & 0xFF
        len_high = report_len >> 8

        unique_readings = [x.reading_id for x in readings if x.reading_id != IOTileReading.InvalidReadingID]
        if len(unique_readings) > 0:
            lowest_id = min(unique_readings)
            highest_id = max(unique_readings)

        header = struct.pack("<BBHLLLBBH", cls.ReportType, len_low, len_high, uuid, report_id, sent_timestamp, signature_method, streamer, selector)
        header = bytearray(header)

        packed_readings = bytearray()

        for reading in readings:
            packed_reading = struct.pack("<HHLLL", reading.stream, 0, reading.reading_id, reading.raw_time, reading.value)
            packed_readings += bytearray(packed_reading)

        footer_stats = struct.pack("<LL", lowest_id, highest_id)
        signed_data = header + packed_readings + footer_stats

        if signer is None:
            signer = ChainedAuthProvider()

        try:
            signature = signer.sign(uuid, signature_method, signed_data)
        except NotFoundError:
            raise ExternalError("Could not sign report because no AuthProvider supported the requested signature method for the requested device", device_id=uuid, signature_method=signature_method)

        footer = struct.pack("16s", str(signature['signature'][:16]))
        footer = bytearray(footer)

        data = signed_data + footer
        return SignedListReport(data)

    def decode(self):
        """Decode this report into a list of readings
        """

        fmt, len_low, len_high, device_id, report_id, sent_timestamp, signature_flags, origin_streamer, streamer_selector = unpack("<BBHLLLBBH", self.raw_report[:20])

        assert fmt == 1
        length = (len_high << 8) | len_low

        self.origin = device_id
        self.report_id = report_id
        self.sent_timestamp = sent_timestamp
        self.origin_streamer = origin_streamer
        self.streamer_selector = streamer_selector
        self.signature_flags = signature_flags

        assert len(self.raw_report) == length

        remaining = self.raw_report[20:]
        assert len(remaining) >= 24
        readings = remaining[:-24]
        footer = remaining[-24:]

        lowest_id, highest_id, signature = unpack("<LL16s", footer)
        signature = bytearray(signature)

        signed_data = self.raw_report[:-16]
        signer = ChainedAuthProvider()

        try:
            verification = signer.verify(device_id, signature_flags, signed_data, signature)
            self.verified = verification['verified']
        except NotFoundError:
            pass

        self.signature = signature
        self.lowest_id = lowest_id
        self.highest_id = highest_id

        #Make sure this report has an integer number of readings
        assert (len(readings) % 16) == 0

        time_base = self.received_time - datetime.timedelta(seconds=sent_timestamp)
        parsed_readings = []

        for i in xrange(0, len(readings), 16):
            reading = readings[i:i+16]
            stream, _, reading_id, timestamp, value = unpack("<HHLLL", reading)

            parsed = IOTileReading(timestamp, stream, value, time_base=time_base, reading_id=reading_id)
            parsed_readings.append(parsed)

        #FIXME: attempt to validate the signature
        return parsed_readings

    def encode(self):
        """Turn this report into a serialized bytearray that could be decoded with a call to decode

        If we were generated from a binary report, just return that so that the signature remains intact,
        otherwise generate one and attempt to sign it
        """

        if hasattr(self, 'raw_report'):
            return self.raw_report

        lowest_id = IOTileReading.InvalidReadingID
        highest_id = IOTileReading.InvalidReadingID

        report_len = 20 + 16*len(self.visible_readings) + 24
        len_low = report_len & 0xFF
        len_high = report_len >> 8

        if len(self.visible_readings) > 0:
            lowest_id = min([x.reading_id for x in self.visible_readings if x.reading_id != IOTileReading.InvalidReadingID])
            highest_id = max([x.reading_id for x in self.visible_readings if x.reading_id != IOTileReading.InvalidReadingID])

        header = struct.pack("<BBHLLLBBH", self.ReportType, len_low, len_high, self.origin, self.report_id, self.sent_timestamp, self.signature_flags, self.origin_streamer, self.streamer_selector)
        header = bytearray(header)

        readings = bytearray()

        for reading in self.visible_readings:
            packed_reading = struct.pack("<HHLLL", reading.stream, 0, reading.reading_id, reading.raw_time, reading.value)
            readings += bytearray(packed_reading)

        #FIXME: Actually calculate a footer here
        footer = struct.pack("<LL16s", lowest_id, highest_id, '\0'*16)
        footer = bytearray(footer)

        return header + readings + footer



"""Base class for data streamed from an IOTile device
"""

import datetime
from iotile.core.exceptions import NotFoundError, ArgumentError

class IOTileReading(object):
    """Base class for readings streamed from IOTile device.

    Each reading represents a single time/value pair sent from an IOTile Device.
    Since many IOTile Devices do not have a hardware realtime clock, the timestamp
    that is assigned to a reading may only be a relative interval from a fixed
    event in the past, like the time the device turned on.

    If the user knows the absolute time for this event they can pass it as a datetime
    in time_base to turn the relative reading timestamp into an absolute datetime
    accessible as reading_time.

    Args:
        raw_time (int): the number of seconds since the device turned on
            when the reading was taken
        time_base (datetime): An optional estimate of when the device was
            last turned on so that we can calulate the actual time of the
            reading
        reading_id (int): An optional unique identifier for this reading that allows
            deduplication.  If no reading id is passed, InvalidReadingID is used.
        stream (int): The stream that this reading is part of
        value (int): The raw reading value
    """

    InvalidReadingID = 0

    def __init__(self, raw_time, stream, value, time_base=None, reading_id=None):
        self.raw_time = raw_time
        self.stream = stream
        self.value = value

        if reading_id is None:
            reading_id = IOTileReading.InvalidReadingID

        self.reading_id = reading_id

        self.reading_time = None
        if time_base is not None:
            self.reading_time = time_base + datetime.timedelta(seconds=raw_time)

    def __eq__(self, other):
        return self.raw_time == other.raw_time and self.stream == other.stream and self.value == other.value and self.reading_id == other.reading_id

    def __str__(self):
        if self.reading_time is not None:
            return "Stream {}: {} at {}".format(self.stream, self.value, self.reading_time)
        else:
            return "Stream {}: {} at uncorrected time {}".format(self.stream, self.value, self.raw_time)

class IOTileReport(object):
    """Base class for data streamed from an IOTile device.

    All IOTileReports must derive from this class and must implement the following interface

    - class method HeaderLength(cls)
        function returns the number of bytes that must be read before the total length of
        the report can be determined.  HeaderLength() must always be less than or equal to
        the length of the smallest version of this report.
    - class method ReportLength(cls, header):
        function that takes HeaderLength() bytes and returns the total size of the report,
        including the header.
    - class method FromReadings(cls, uuid, readings)
        function that creates an instance of an IOTileReport subclass from a list of readings
        and a device uuid.
    - propery ReportType:
        The one byte type code that defines this report type
    - instance method verify(self):
        function that verifies that a report is correctly received and, if possible, that
        the sender is who it says it is.
    - instance method decode(self):
        function that decodes a report into a series of IOTileReading objects. The function
        should return a list of readings.
    - instance method serialize(self):
        function that should turn the report into a serialized bytearray that could be
        decoded with decode().

    Args:
        rawreport (bytearray): The raw data of this report
        signed (bool): Whether this report is signed to specify who it is from
        encrypted (bool): Whether this report is encrypted\
        received_time (datetime): The time in UTC when this report was received from a device.
            If not received, the time is assumed to be utcnow().
    """

    def __init__(self, rawreport, signed, encrypted, received_time=None):
        self.visible_readings = []
        self.origin = None

        if received_time is None:
            self.received_time = datetime.datetime.utcnow()
        else:
            self.received_time = received_time

        self.raw_report = rawreport
        self.signed = signed
        self.encrypted = encrypted
        self.decoded = False
        self.verified = False

        # If we're able to, decode the report immediately
        if not encrypted:
            self.visible_readings = self.decode()
            self.decoded = True

    @classmethod
    def HeaderLength(cls):
        """Return the length of a header needed to calculate this report's length

        Returns:
            int: the length of the needed report header
        """

        raise NotFoundError("IOTileReport HeaderLength needs to be overridden")

    @classmethod
    def ReportLength(cls, header):
        """Given a header of HeaderLength bytes, calculate the size of this report
        """

        raise NotFoundError("IOTileReport ReportLength needs to be overriden")

    def decode(self):
        """Decode a raw report into a series of readings
        """

        raise NotFoundError("IOTileReport decode needs to be overriden")

    def encode(self):
        """Encode this report into a binary blob that could be decoded by a report format's decode method
        """

        return self.raw_report

    def save(self, path):
        """Save a binary copy of this report

        Args:
            path (string): The path where we should save the binary copy of the report
        """

        data = self.encode()

        with open(path, "wb") as out:
            out.write(data)

    def serialize(self):
        """Turn this report into a dictionary that encodes all information including received timestamp
        """

        info = {}
        info['received_time'] = self.received_time
        info['encoded_report'] = str(self.encode())
        info['report_format'] = ord(info['encoded_report'][0])  # Report format is the first byte of the encoded report
        info['origin'] = self.origin

        return info

    def __str__(self):
        return "IOTile Report of length %d with %d visible readings" % (len(self.raw_report), len(self.visible_readings))



"""An ordered list of authentication providers that are checked in turn to attempt a crypto operation
"""

import pkg_resources
from .auth_provider import AuthProvider, KnownSignatureMethods
from iotile.core.exceptions import NotFoundError, ExternalError

class ChainedAuthProvider(AuthProvider):
    """An AuthProvider that delegates operations to a list of subproviders

    Each subprovider is checked in turn and the first one able to satisfy a request
    is used.  If no provider is available that can perform an operation, NotFoundError
    is raised.

    By default, the python entry_point group 'iotile.default_auth_providers' is used
    to build the list of auth providers in the chain.  The entries in that group should
    be tuples of (priority, auth_provider_class, arg_dict) where priority is an integer,
    auth_provider_class is an AuthProvider subclass and arg_dict is a dictionary of
    arguments passed to the constructor of auth_provider.
    """

    def __init__(self, args=None):
        #FIXME: Allow overwriting default providers via args
        self._load_installed_providers()

        sub_providers = []
        for entry in pkg_resources.iter_entry_points('iotile.default_auth_providers'):
            priority, provider, args = entry.load()

            if provider not in self._auth_factories:
                raise ExternalError("Default authentication provider list references unknown auth provider", provider_name=provider, known_providers=self._auth_factories.keys())
            configured = self._auth_factories[provider](args)
            sub_providers.append((priority, configured))

        sub_providers.sort(key=lambda x: x[0])
        self.providers = sub_providers

    def _load_installed_providers(self):
        self._auth_factories = {}

        for entry in pkg_resources.iter_entry_points('iotile.auth_provider'):
            self._auth_factories[entry.name] = entry.load()

    def encrypt(self, device_id, enc_method, data):
        """Encrypt a buffer of data on behalf of a device

        Args:
            device_id (int): The id of the device that we should encrypt for
            enc_method (int): The method of encryption that we should perform
            data (bytearray): The data that we should encrypt

        Returns:
            dict: The encrypted data and any associated metadata about the data.
                The data itself must always be a bytearray stored under the 'data'
                key, however additional keys may be present depending on the encryption method
                used.

        Raises:
            NotFoundError: If the auth provider is not able to encrypt the data.
        """

        for priority, provider in self.providers:
            try:
                return provider.encrypt(device_id, enc_method, data)
            except NotFoundError:
                pass

        raise NotFoundError("encrypt method is not implemented in any sub_providers")

    def decrypt(self, device_id, enc_method, data):
        """Decrypt a buffer of data on behalf of a device

        Args:
            device_id (int): The id of the device that we should encrypt for
            enc_method (int): The method of encryption that we should perform
            data (bytearray): The data that we should encrypt

        Returns:
            dict: The decrypted data and any associated metadata about the data.
                The data itself must always be a bytearray stored under the 'data'
                key, however additional keys may be present depending on the encryption method
                used.

        Raises:
            NotFoundError: If the auth provider is not able to encrypt the data.
        """

        for priority, provider in self.providers:
            try:
                return provider.decrypt(device_id, enc_method, data)
            except NotFoundError:
                pass

        raise NotFoundError("decrypt method is not implemented in any sub_providers")

    def sign(self, device_id, sig_method, data):
        """Sign a buffer of data on behalf of a device

        Args:
            device_id (int): The id of the device that we should encrypt for
            sig_method (int): The method of encryption that we should perform
            data (bytearray): The data that we should sign

        Returns:
            dict: The signature and any associated metadata about the signature.
                The signatured itself must always be a bytearray stored under the
                'signature' key, however additional keys may be present depending
                on the signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to encrypt the data.
        """

        for priority, provider in self.providers:
            try:
                return provider.sign(device_id, sig_method, data)
            except NotFoundError:
                pass

        raise NotFoundError("sign method is not implemented in any sub_providers")

    def verify(self, device_id, sig_method, data, signature):
        """Verify the signature attached to a buffer of data

        Args:
            device_id (int): The id of the device that we should encrypt for
            sig_method (int): The method of signing that was used
            data (bytearray): The data whose signature we should verify
            signature (bytearray): The signature attached to data

        Returns:
            dict: The result of the verification process must always be a bool under the
                'verified' key, however additional keys may be present depending on the
                signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to encrypt the data.
        """

        for priority, provider in self.providers:
            try:
                return provider.verify(device_id, sig_method, data, signature)
            except NotFoundError:
                pass

        raise NotFoundError("verify method is not implemented in any sub_providers")



import pytest
import threading
import sys

if sys.version[0] == '2':
    import Queue as queue
else:
    import queue as queue

from iotile.core.utilities.stoppable_thread import StoppableWorkerThread
from iotile.core.exceptions import TimeoutExpiredError

def test_running_function():
    """Make sure we can run a function in the thread
    """

    items = queue.Queue()

    def thread_func(shared_queue):
        if shared_queue.qsize() == 3:
            return

        shared_queue.put(1)

    thread = StoppableWorkerThread(thread_func, timeout=0.001, args=(items,))
    thread.start()

    items.get(block=True, timeout=1.0)
    items.get(block=True, timeout=1.0)
    items.get(block=True, timeout=1.0)

    thread.stop()

def test_kwargs_function():
    """Make sure we can run a function in the thread with kwargs
    """

    items = queue.Queue()

    def thread_func(shared_queue):
        if shared_queue.qsize() == 3:
            return

        shared_queue.put(1)

    thread = StoppableWorkerThread(thread_func, timeout=0.001, kwargs={'shared_queue': items})
    thread.start()

    items.get(block=True, timeout=1.0)
    items.get(block=True, timeout=1.0)
    items.get(block=True, timeout=1.0)

    thread.stop()

def test_running_generator():
    """Make sure we can run a generator in the thread
    """

    items = queue.Queue()

    def thread_func(shared_queue):
        for i in xrange(0, 3):
            shared_queue.put(i)
            yield

    thread = StoppableWorkerThread(thread_func, timeout=0.001, args=(items,))
    thread.start()

    a = items.get(block=True, timeout=1.0)
    b = items.get(block=True, timeout=1.0)
    c = items.get(block=True, timeout=1.0)

    assert a == 0
    assert b == 1
    assert c == 2

    thread.stop()

def test_failed_stop():
    """Make sure that we can always kill a thread if it doesn't respond to stop()
    """

    def thread_func():
        while True:
            pass

    thread = StoppableWorkerThread(thread_func, timeout=0.001)
    thread.start()
    thread.wait_running(timeout=1.0)

    with pytest.raises(TimeoutExpiredError):
        thread.stop(timeout=0.01)

def test_failed_stop_force():
    """Make sure that we can always kill a thread if it doesn't respond to stop()
    """

    def thread_func():
        while True:
            pass

    thread = StoppableWorkerThread(thread_func, timeout=0.001)
    thread.start()

    thread.stop(timeout=0.01, force=True)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

# Copyright (c) 2011, Bernhard Leiner
# All rights reserved.
#
# Redistribution and use in source and binary forms,
# with or without modification, are permitted provided
# that the following conditions are met:
#
# * Redistributions of source code must retain
#   the above copyright notice, this list of conditions
#   and the following disclaimer.
# * Redistributions in binary form must reproduce
#   the above copyright notice, this list of conditions
#   and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
# * Neither the name of the author nor the names
#   of its contributors may be used to endorse
#   or promote products derived from this software
#   without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

'''Compatibility functions for python 2 and 3.

@author     Bernhard Leiner (bleiner AT gmail com)
@version    1.0
'''

__docformat__ = "javadoc"


import sys

if sys.version_info[0] >= 3:
    def asbytes(s):
        if isinstance(s, bytes):
            return s
        return s.encode('latin1')
    def asstr(s):
        if isinstance(s, str):
            return s
        return s.decode('latin1')
else:
    asbytes = str
    asstr = str

